"""
Visualize a dataset's data model as entity-relationship diagram of the corresponding CLDF SQL.
"""
import shlex
import shutil
import pathlib
import tempfile
import subprocess

import requests
from pycldf.ext.sql import get_database
from clldutils.clilib import PathType
from clldutils.path import ensure_cmd

from cldfviz.cli_util import add_testable, add_open, write_output


def download_file(url, target):
    with requests.get('https://github.com/' + url, stream=True) as r:
        r.raise_for_status()
        with target.open('wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return target


def copy_file(dest, target):
    shutil.copy(dest, target)
    return target


def register(parser):
    add_testable(parser)
    parser.add_argument('dataset_locator')
    parser.add_argument(
        '--java',
        type=ensure_cmd,
        help='Path to the Java runtime.',
        default='java')
    parser.add_argument(
        '--schemaspy-jar',
        type=PathType(type='file'),
        help='Path to a suitable version of the SchemaSpy jar file.',
        default=None)
    parser.add_argument(
        '--sqlite-jar',
        type=PathType(type='file'),
        help='Path to a suitable version of the Xerial SQLite JDBC Driver jar file.',
        default=None)
    parser.add_argument(
        '--db',
        type=PathType(type='file'),
        help='Path to CLDF SQLite file. This provides a shortcut, bypassing SQLite creation.',
        default=None,
    )
    parser.add_argument(
        '--format',
        choices=['compact.dot', 'compact.svg', 'large.dot', 'large.svg'],
        help="`large` diagrams include all fields of an entity, `compact` ones do not. Diagrams "
             "are available in SVG or Graphviz' DOT language.",
        default='large.svg')
    add_open(parser)


def run(args):
    with tempfile.TemporaryDirectory() as tmp:
        tmp = pathlib.Path(tmp)
        for url, target in [
            ('xerial/sqlite-jdbc/releases/download/3.39.4.1/sqlite-jdbc-3.39.4.1.jar',
             'sqlite.jar'),
            ('schemaspy/schemaspy/releases/download/v6.1.0/schemaspy-6.1.0.jar',
             'schemaspy.jar'),
        ]:
            attrib = target.replace('.', '_')
            if not getattr(args, attrib):
                setattr(args, attrib, download_file(url, tmp / target))
            else:
                setattr(args, attrib, copy_file(getattr(args, attrib), tmp / target))
        if args.db:
            copy_file(args.db, tmp / 'db.sqlite')  # pragma: no cover
        else:
            get_database(args.dataset_locator, download_dir=tmp, fname=tmp / 'db.sqlite')

        # Note: This exact way of calling java must be kept to keep tests working.
        out = subprocess.check_output(shlex.split(
            "{} -jar {} -t sqlite-xerial -db {} -sso -s public -dp {} -o {} -cat % -vizjs".format(
                args.java,
                args.schemaspy_jar,
                tmp / 'db.sqlite',
                tmp,
                args.sqlite_jar.parent,
            )),
        )
        for line in out.decode('utf8').split('\n'):  # pragma: no cover
            args.log.debug(line)
        res = tmp\
            .joinpath('diagrams', 'summary', 'relationships.real.{}'.format(args.format))\
            .read_text(encoding='utf8')

    write_output(args, res)



import argparse

import pycldf

from . import leaflet  # noqa: F401
try:
    from . import mpl  # noqa: F401
    WITH_CARTOPY = True
except ImportError:  # pragma: no cover
    WITH_CARTOPY = False
    mpl = None
from .base import Map  # noqa: F401


class MarkerFactory:
    """
    If the capabilities of standard `Map` implementations are not sufficient, custom plotting
    code can be provided as subclass of `MarkerFactory`.

    Note:
    It's the responsibility of the `MarkerFactory` to figure out what map format is requested.
    This can be done by inspecting `self.args.format`, or checking the type of `map` using
    `isinstance` in the drawing methods.
    """
    def __init__(self, dataset: pycldf.Dataset, args: argparse.Namespace, *cfg):
        """
        Called when the CLDF data is available.

        :param dataset: The CLDF data.
        :param args: The cli arguments.
        :param cfg: Whatever has been passed to `--marker-factory` in addition to the module \
        providing the `MarkerFactory` (split by comma).
        """
        self.dataset = dataset
        self.args = args
        self.cfg = cfg

    def __call__(self, map: Map, language, values, colormaps):
        """
        Called for each language on the map. An implementation must return either
        - `True`: to signal that all plotting has been done (e.g. by plotting directly to `map.ax` \
          in the case of matplotlib) or
        - an instance of `map.__marker_class__`, providing a marker specification to `map`.

        :param map:
        :param language:
        :param values:
        :param colormaps:
        :return:
        """
        if self.args.format == 'html':
            return leaflet.LeafletMarkerSpec()
        return mpl.MPLMarkerSpec()

    def legend(self, map, parameters, colormaps):
        """
        :param map:
        :param parameters:
        :param colormaps:
        :return:
        """
        return


class _TestFactory(MarkerFactory):  # Only used for testing import from module.
    pass



import argparse

import pytest
from pycldf import Dataset

from cldfviz.multiparameter import MultiParameter, Language, Value, CONTINUOUS
from cldfviz.cli_util import get_language_filter


def test_Language(glottolog, StructureDataset):
    from pycldf import orm
    for row in StructureDataset['LanguageTable']:
        row['Glottocode'] = 'abcd1234'
        row['Latitude'] = None
        lang = Language(orm.Language(StructureDataset, row), glottolog)
        assert lang.lat == pytest.approx(10.0)
        break

    with pytest.raises(TypeError):
        _ = Language({})


def test_MultiParameter(metadatafree_dataset, StructureDataset, glottolog, tmp_path):
    with pytest.raises(ValueError):
        _ = MultiParameter(StructureDataset, ['A'])

    _ = MultiParameter(metadatafree_dataset, ['param1'], glottolog=glottolog)
    mp = MultiParameter(StructureDataset, ['B', 'C'])
    for lang, values in mp.iter_languages():
        assert lang.name == 'Bengali'
        assert values['C'][0].v == 'C-1'
        assert values['C'][0].code == '1'
        break
    mp = MultiParameter(
        StructureDataset,
        ['B'],
        language_filter=get_language_filter(argparse.Namespace(language_filters='{"Filtered":"False"}')),
        language_properties=['Family_name'])
    assert len(mp.languages) == 26
    assert 'Family_name' in mp.parameters
    mp = MultiParameter(StructureDataset, [])
    assert '__language__' in mp.parameters

    values = tmp_path / 'values.csv'
    values.write_text("""\
ID,Language_ID,Parameter_ID,Value
1,abcd1235,param1,1
2,abcd1235,param1,2
3,abcd1235,param1,3
4,abcd1235,param1,4
5,abcd1235,param1,5
6,abcd1235,param1,6
7,abcd1235,param1,7
8,abcd1235,param1,8
9,abcd1235,param1,9
10,abcd1234,param1,10""", encoding='utf8')
    ds = Dataset.from_data(values)
    mp = MultiParameter(ds, ['param1'], glottolog=glottolog)
    assert list(mp.parameters.values())[0].type == CONTINUOUS
    mp = MultiParameter(ds, [], glottolog=glottolog)
    assert len(mp.languages) == 2


def test_Value():
    v1 = Value(v=1, pid=1, lid=1, code=1)
    assert v1 == Value(v=1, pid=1, lid=1, code=2)
    assert Value(v=2, pid=2, lid=1, code=2) < Value(v=1, pid=1, lid=2, code=1)



"""
Build an HTML file to display a wordlist for a particular concept with associated audio.

Assumes that
- the dataset is a CLDF Wordlist, coding concepts as parameters,
- forms are linked to media files
  - via a column with propertyUrl "mediaReference" in FormTable or
  - a column with propertyUrl "formReference" in MediaTable.
"""
from clldutils.clilib import PathType
from clldutils.misc import nfilter
from pycldf.terms import term_uri
from pycldf.cli_util import get_dataset, add_dataset

from cldfviz.cli_util import add_open, write_output, add_jinja_template
from cldfviz.media import get_objects_and_media, get_media_url
from cldfviz.template import render_jinja_template, TEMPLATE_DIR


def register(parser):
    add_dataset(parser)
    parser.add_argument(
        'concept',
        help='Specify a concept in the form COLNAME=VALUE or cldf:PROPNAME=VALUE, e.g. '
             '"Name=hand" or "cldf:id=10", using column names used in ParameterTable or '
             'CLDF properties available in ParameterTable.')
    # parser.add_argument('--mimetype', default=None)
    mod = __name__.split('.')[-1]
    add_jinja_template(parser, TEMPLATE_DIR / mod / '{}.html'.format(mod))
    parser.add_argument(
        '--media-dir',
        help="If media files are available locally (downloaded via pycldf's `cldf downloadmedia` "
             "command), the media directory can be passed to access the audio files on disk.",
        type=PathType(type='dir'),
        default=None)
    add_open(parser)


def run(args):
    ds = get_dataset(args)

    # Determine the relevant concept (aka parameter):
    colspec, sep, match = args.concept.partition('=')
    colname = None
    if not sep:
        match = colspec
    else:
        cldfprop = colspec.startswith('cldf:')
        colspec = colspec.replace('cldf:', '')
        colname = ds['ParameterTable', term_uri(colspec)].name if cldfprop else colspec

    if 'ParameterTable' in ds:
        for row in ds.iter_rows('ParameterTable', 'id', 'name'):
            if (colname and row[colname] == match) or row['id'] == match:
                pid = row['id']
                break
        else:
            raise ValueError(args.concept)  # pragma: no cover
    else:
        assert colname is None
        pid = match

    # Get relevant forms and linked media:
    forms = []
    for form, media in get_objects_and_media(
            ds, 'FormTable', 'formReference', filter=lambda f: f.cldf.parameterReference == pid):
        forms.append((
            form,
            nfilter(get_media_url(f, args.media_dir) for f in media if f.mimetype.type == 'audio')))

    res = render_jinja_template(
        args.template,
        ds=ds,
        pid=pid,
        parameter=forms[0][0].parameter,
        forms=forms,
        local=bool(args.media_dir))
    write_output(args, res)



import warnings

import pytest


@pytest.mark.parametrize(
    'opts,expect',
    [
        (
            dict(
                glottolog_mapping={
                    'Santali_NM': ('abcd1234', 'Abcd'),
                    'Marathi_IA': ('efgh1234', None),
                },
                legend='The Tree',
                with_glottolog_links=True),
            lambda svg: 'The Tree' in svg and 'Abcd' in svg),
        (
            dict(labels=lambda n: '(' + n.name + ')'),
            lambda svg: '_Santali_NM_' in svg),
        (
            dict(leafs=lambda n: n.name != 'Santali_NM'),
            lambda svg: 'Santali_NM' not in svg),
    ]
)
def test_render(StructureDataset, opts, expect):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=DeprecationWarning, module='importlib._bootstrap')
        from pycldf.trees import TreeTable
        from cldfviz.tree import render
        from newick import loads

    svg = render(list(TreeTable(StructureDataset))[0], **opts)
    assert '<svg' in svg
    assert expect(svg)

    render(loads('(A,);')[0], labels={'A': None})


def test_render_to_file(StructureDataset, tmp_path):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=DeprecationWarning, module='importlib._bootstrap')
        from pycldf.trees import TreeTable
        from cldfviz.tree import render

    o = tmp_path / 'test.svg'
    render(list(TreeTable(StructureDataset))[0], output=o)
    assert o.exists()



import argparse

import pytest

from cldfviz.glottolog import *


@pytest.fixture
def glottolog_cldf(glottolog_dir, tmp_path):
    from pycldf import StructureDataset
    import pyglottolog

    ds = StructureDataset.in_dir(tmp_path / 'glottolog-cldf')
    ds.add_component('LanguageTable')
    langs, vals = [], []
    for lang in pyglottolog.Glottolog(glottolog_dir).languoids():
        langs.append(dict(
            ID=lang.id, Name=lang.name, Latitude=lang.latitude, Longitude=lang.longitude))
        vals.append(dict(
            ID=lang.id,
            Language_ID=lang.id,
            Parameter_ID='subclassification',
            Value=lang.newick_node(template="{l.id}").newick,
        ))
    ds.write(LanguageTable=langs, ValueTable=vals)
    return tmp_path / 'glottolog-cldf'


def test_Glottolog_glottolog_cldf(glottolog_cldf):
    parser = argparse.ArgumentParser()
    Glottolog.add(parser)
    args = parser.parse_args(['--glottolog-cldf', str(glottolog_cldf)])
    gl = Glottolog.from_args(args)
    assert 'abcd1234' in gl
    assert len(gl) == 8
    assert gl.newick('abcd1234')


def test_Glottolog_pyglottolog(glottolog_dir):
    parser = argparse.ArgumentParser()
    Glottolog.add(parser)
    args = parser.parse_args(['--glottolog', str(glottolog_dir)])
    gl = Glottolog.from_args(args)
    assert 'abcd1234' in gl
    assert len(gl) == 8
    assert gl.newick('abcd1234')



import re
import html
import typing
import pathlib
import functools

from pycldf import Dataset
from pycldf.ext.markdown import CLDFMarkdownText
import jinja2
import jinja2.meta
from clldutils.misc import nfilter
from clldutils.markup import MarkdownLink, MarkdownImageLink

import cldfviz

__all__ = ['iter_templates', 'render', 'iter_cldfviz_links']

TEMPLATE_DIR = cldfviz.PKG_DIR.joinpath('templates', 'text')


def _add_filters(env):
    def paragraphs(s):
        return '\n\n'.join(s.split('\n'))

    env.filters['paragraphs'] = paragraphs


def get_env(template_dir=None, fallback_template_dir=None):
    loader = jinja2.FileSystemLoader(
        searchpath=[str(d) for d in nfilter([template_dir, fallback_template_dir, TEMPLATE_DIR])])
    env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
    _add_filters(env)
    return env


def iter_templates():
    env = get_env()
    for p in sorted(TEMPLATE_DIR.iterdir(), key=lambda pp: pp.name):
        m = re.match(r"{#(.+?)#}", p.read_text(encoding='utf8'), flags=re.MULTILINE | re.DOTALL)
        doc = m.group(1) if m else None
        parsed_content = env.parse(env.loader.get_source(env, p.name))
        vars = jinja2.meta.find_undeclared_variables(parsed_content)
        yield p, doc, [v for v in vars if v != 'ctx']


def pad_ex(obj: typing.Iterable[str],
           gloss: typing.Iterable[str],
           escape: typing.Optional[bool] = True):
    """
    :param escape: Flag signaling whether to html.escape words and glosses.
    """
    out_obj = []
    out_gloss = []
    for o, g in zip(obj, gloss):
        g = g or ""
        diff = len(o) - len(g)
        if diff < 0:
            o += " "*-diff  # noqa E225
        else:
            g += " " * diff
        out_obj.append(html.escape(o, quote=False) if escape else o)
        out_gloss.append(html.escape(g, quote=False) if escape else g)
    return "  ".join(out_obj).strip(), "  ".join(out_gloss).strip()


def render(doc: typing.Union[pathlib.Path, str],
           cldf_dict: typing.Union[Dataset, typing.Dict[typing.Union[str, None], Dataset]],
           template_dir: typing.Optional[typing.Union[str, pathlib.Path]] = None,
           loader: typing.Optional[jinja2.BaseLoader] = None,
           func_dict: typing.Optional[typing.Dict[str, callable]] = None,
           escape: typing.Optional[bool] = True) -> str:
    """
    Render CLDF Markdown using customizable jinja2 templates.

    Features:

    - Reference list: Include a list of cited references using the link \
      `[](Source?cited_only#cldf:__all__)`

    :param doc: A CLDF Markdown document specified as string or filepath.
    :param cldf_dict: A CLDF dataset or a mapping of prefixes to CLDF datasets.
    :param template_dir: Path to custom template directory.
    :param loader: As alternative to a custom template directory, a custom jinja2 loader can be \
    specified.
    :param func_dict: Mapping of names to callables passed to templates as renderer globals, see \
    https://jinja.palletsprojects.com/en/3.1.x/api/#jinja2.Environment.globals.
    :return: Rendered document as string.
    """
    func_dict = func_dict or {}
    func_dict.update({"pad_ex": functools.partial(pad_ex, escape=escape)})

    if isinstance(cldf_dict, Dataset):
        cldf_dict = {None: cldf_dict}

    folder_env = get_env(template_dir=template_dir)
    if loader is None:
        env = folder_env
    else:
        env = jinja2.Environment(
            loader=jinja2.ChoiceLoader([loader, folder_env.loader]),
            trim_blocks=True,
            lstrip_blocks=True)
        _add_filters(env)

    proc = TemplateRenderer(
        env, func_dict, doc, cldf_dict,
        # download_dir!
    )
    return proc.render()


def iter_cldfviz_links(md):
    for match in MarkdownImageLink.pattern.finditer(md):
        ml = MarkdownImageLink.from_match(match)
        if re.match(r'cldfviz\.(map|tree)', ml.parsed_url.fragment):
            yield ml


class TemplateRenderer(CLDFMarkdownText):
    def __init__(self, env, func_dict, *args, **kw):
        super().__init__(*args, **kw)
        self.env = env
        self.func_dict = func_dict
        self.with_partial_local_reflist = False
        self.cited = None

    def render_template(self, fname_or_component, ctx, index=False, fmt='md'):
        # Determine the template to use ...
        tmpl_fname = ctx.pop(
            '__template__',  # ... by looking for an explicit name ...
            # ... and falling back to the "most suitable" one.
            '{}_{}.{}'.format(fname_or_component, 'index' if index else 'detail', fmt),
        )
        jinja_template = self.env.get_template(tmpl_fname)
        jinja_template.globals.update(self.func_dict)
        jinja_template.globals.update({"component": fname_or_component})
        return jinja_template.render(**ctx)

    def get_tmpl_context(self, ml):
        tmpl_context = {k: True if v[0] == '' else v[0] for k, v in ml.parsed_url_query.items()}
        for k in tmpl_context:
            if k.startswith('with_') and (tmpl_context[k] in ['0', 'false', 'False']):
                tmpl_context[k] = False
        tmpl_context['ctx'] = self.get_object(ml)
        tmpl_context['cldf'] = self.dataset_mapping[ml.prefix]
        return tmpl_context

    def render_link(self, ml):
        ref_link = ml.all and \
            'cited_only' in ml.parsed_url_query and \
            ml.component(self.dataset_mapping[ml.prefix]) == 'Source'

        if self.cited:
            # We're in the second pass!
            if ref_link:
                ctx = self.get_tmpl_context(ml)
                ctx['ctx'] = [s for s in self.get_object(ml) if s.id in self.cited]
                ctx["with_anchor"] = True
                return self.render_template('Source', ctx, index=True)
            return ml

        if ref_link:  # So we need a second pass.
            self.with_partial_local_reflist = True
            return ml

        return self.render_template(
            ml.component(self.dataset_mapping[ml.prefix]) or ml.table_or_fname,
            self.get_tmpl_context(ml),
            index=ml.all)

    def render(self, simple_link_detection=True, markdown_kw=None):
        md = super().render(simple_link_detection=simple_link_detection, markdown_kw=markdown_kw)
        if self.with_partial_local_reflist:
            # Second pass!
            # 1. Determine which sources have been referenced:
            self.cited = set(
                MarkdownLink.from_match(ml).url.split('-', maxsplit=1)[-1]
                for ml in MarkdownLink.pattern.finditer(md)
                if MarkdownLink.from_match(ml).url.startswith('#source-'))

            # 2. Insert the pruned list of sources:
            md = super().render(
                simple_link_detection=simple_link_detection, markdown_kw=markdown_kw)
        return md



import json
import typing
import itertools
import collections

from matplotlib import cm
from matplotlib.colors import Normalize, to_hex, CSS4_COLORS, BASE_COLORS
import matplotlib.pyplot as plt
from clldutils.color import qualitative_colors, sequential_colors, rgb_as_hex

from cldfviz.multiparameter import CONTINUOUS, CATEGORICAL, Parameter

__all__ = ['COLORMAPS', 'hextriplet', 'Colormap', 'get_shape_and_color', 'weighted_colors']
COLORMAPS = {
    CATEGORICAL: ['boynton', 'tol', 'base', 'seq'],
    CONTINUOUS: [cm for cm in plt.colormaps() if not cm.endswith('_r')],
}
SHAPES = {
    'triangle_down',
    'triangle_up',
    'square',
    'diamond',
    'circle',
}
SVG_SHAPE_MAP = {
    'triangle_down': 'f',
    'triangle_up': 't',
    'square': 's',
    'diamond': 'd',
    'circle': 'c',
}


def hextriplet(s):
    """
    Wrap clldutils.color.rgb_as_hex to provide unified error handling.
    """
    if s in SHAPES:
        # A bit of a hack: We allow a handful of shape names as "color" spec as well.
        return s
    if s in BASE_COLORS:
        return rgb_as_hex([float(d) for d in BASE_COLORS[s]])
    if s in CSS4_COLORS:
        return CSS4_COLORS[s]
    try:
        return rgb_as_hex(s)
    except (AssertionError, ValueError) as e:
        raise ValueError('Invalid color spec: "{}" ({})'.format(s, str(e)))


class Colormap:
    def __init__(self, parameter: Parameter, name: typing.Optional[str] = None, novalue=None):
        domain = parameter.domain
        self.explicit_cm = None
        if name and name.startswith('{'):
            if isinstance(parameter.domain, tuple):
                raise ValueError(
                    'Explicit color maps are only supported for categorical parameters')
            self.explicit_cm = collections.OrderedDict()
            raw = json.loads(name, object_pairs_hook=collections.OrderedDict)
            if novalue:
                raw.setdefault('None', novalue)
            label_to_code = {v: k for k, v in parameter.domain.items()}
            for v, c in raw.items():
                if (v not in parameter.value_to_code) and v not in label_to_code:
                    raise ValueError('Colormap value "{}" not in domain {}'.format(
                        v, list(parameter.value_to_code.keys())))
                v = parameter.value_to_code.get(v, label_to_code.get(v))
                self.explicit_cm[v] = hextriplet(c)
            vals = list(parameter.value_to_code)
            if len(vals) > len(self.explicit_cm):
                raise ValueError('Colormap {} does not cover all values {}!'.format(
                    dict(raw), vals))
            name = None
            # reorder the domain of the parameter (and prune it to valid values):
            parameter.domain = collections.OrderedDict(
                (c, l) for c, l in sorted(
                    [i for i in parameter.domain.items() if i[0] in self.explicit_cm],
                    key=lambda i: list(self.explicit_cm.keys()).index(i[0]))
            )
        self.novalue = hextriplet(novalue) if novalue else None
        self._cm = getattr(cm, name or 'yyy', cm.jet)

        if isinstance(domain, tuple):
            assert not self.explicit_cm
            # Initialize matplotlib colormap and normalizer:
            norm = Normalize(domain[0], domain[1])
            self.cm = lambda v: to_hex(self._cm(norm(float(v))))
        else:
            if self.explicit_cm:
                self.cm = lambda v: self.explicit_cm[v]
            else:
                if name == 'seq':
                    colors = sequential_colors(len(domain))
                else:
                    colors = qualitative_colors(len(domain), set=name)
                self.cm = lambda v: dict(zip(domain, colors))[v]

    @property
    def with_shapes(self):
        return self.explicit_cm and any(c in SHAPES for c in self.explicit_cm.values())

    def scalar_mappable(self):
        return cm.ScalarMappable(norm=None, cmap=self._cm)

    def __call__(self, value):
        if value is None:
            return self.novalue
        return self.cm(value)


def get_shape_and_color(colors_or_shapes):
    if 1 <= len(colors_or_shapes) <= 2:
        shapes, colors = [], []
        for _, c in colors_or_shapes:
            (shapes if c in SHAPES else colors).append(c)
        if shapes:
            if len(shapes) > 1:
                raise ValueError('Only one shape can be specified for a marker')
            return shapes[0], colors[0] if colors else '#000000'


def weighted_colors(values, colormaps):
    colors = []
    for pid, vals in values.items():
        cm = colormaps[pid]
        total = sum(1 if vv.weight is None else vv.weight for vv in vals)
        for code, vvs in itertools.groupby(sorted(vals, key=lambda vv: vv.v), lambda vv: vv.v):
            colors.append((
                sum(1 if vv.weight is None else vv.weight for vv in vvs) / total / len(values),
                cm(code)))
    return colors



import pytest

from pycldf import Generic
from jinja2 import BaseLoader, TemplateNotFound

from cldfviz.text import *


def test_iter_templates():
    assert len(list(iter_templates())) > 17


def test_regular_md(StructureDataset):
    assert render('[Glottolog](https://glottolog.org)', StructureDataset) == \
           '[Glottolog](https://glottolog.org)'


def test_custom_templates(StructureDataset, tmp_path):
    tmpl = tmp_path / 'tmpl'
    tmpl.mkdir()
    tmpl.joinpath('LanguageTable_detail.md').write_text(
        '{{ "stuff" | paragraphs }}', encoding='utf8')
    assert 'stuff' in render('[](LanguageTable#cldf:Juang_SM)', StructureDataset, template_dir=tmpl)

    class Loader(BaseLoader):
        def get_source(self, environment, template):
            path = tmpl / template
            return path.read_text(encoding='utf8'), path, lambda: 1
    assert 'stuff' in render('[](LanguageTable#cldf:Juang_SM)', StructureDataset, loader=Loader())


def test_non_standard_table(tmp_path):
    ds = Generic.in_dir(tmp_path)
    ds.add_table('data.csv', 'id', 'name', 'description')
    ds.write(tmp_path / 'md.json', **{'data.csv': [dict(id='1', name='The Name', description='')]})
    tmpl = tmp_path / 'tmpl'
    tmpl.mkdir()
    tmpl.joinpath('data.csv_detail.md').write_text("{{ ctx['name'] }}", encoding='utf8')
    assert 'The Name' in render('[](data.csv#cldf:1)', ds, template_dir=tmpl)


def test_render_example(StructureDataset):
    assert render('[ex](ExampleTable#cldf:igt)', StructureDataset) == """\

> (igt) Ho (Peterson 2017)
<pre>
The  t-e&lt;in&gt;xt  
The  GL-&lt;FOC&gt;OSS  
‘the translation’</pre>
"""
    assert render('[ex](ExampleTable#cldf:igt)', StructureDataset, escape=False) == """\

> (igt) Ho (Peterson 2017)
<pre>
The  t-e<in>xt  
The  GL-<FOC>OSS  
‘the translation’</pre>
"""
    assert render('[ex](ExampleTable#cldf:ptonly)', StructureDataset) == """\

> (ptonly) Ho
<pre>
<i>Only primary text</i>  
‘and translation’</pre>
"""
    res = render(
        """[](ExampleTable?with_internal_ref_link#cldf:igt)
[](sources.bib#cldf:__all__)""",
        StructureDataset)
    assert "The Title" in res
    res = render(
        """[](ExampleTable?with_internal_ref_link#cldf:igt)
[](sources.bib?cited_only=1#cldf:__all__)""",
        StructureDataset)
    assert "The Title" not in res
    assert "Fitting the pieces" in res


def test_render_entry(Dictionary):
    assert render('[ex](EntryTable#cldf:entry-1)', Dictionary) \
        == '‘Motschegiebschn’, *noun*: lady bug\n'
    assert render('[ex](EntryTable#cldf:entry-2)', Dictionary) \
        == '‘Bemme’, *noun*: 1. sandwich 2. tire of a bicycle\n'
    assert render('[ex](EntryTable#cldf-pref:entry-2)', dict(pref=Dictionary)) \
           == '‘Bemme’, *noun*: 1. sandwich 2. tire of a bicycle\n'


def test_render_metadata(StructureDataset):
    assert render('[](p/StructureDataset-metadata.json#cldf:"dc:identifier")', StructureDataset) == \
        'https://doi.org/10.1515/jsall-2017-0008'
    assert render(
        """[](Metadata#cldf:tables[?"dc:conformsTo"=='http://cldf.clld.org/v1.0/terms.rdf#LanguageTable'].url | [0])""",
        StructureDataset) == 'languages.csv'


@pytest.mark.parametrize(
    'comp,query,oid,ds,expected',
    [
        ('BorrowingTable', None, 'dutch_eiland', 'wl', 'Old Frisian'),
        ('CodeTable', None, 'C-0', 'sd', 'affixal'),
        ('CognateTable', None, 'Tena-1_one-1-1', 'wl', 'List and'),
        ('CognateTable', 'cognatesetReference=1', '__all__', 'wl', 'hʊk'),
        ('CognatesetTable', None, '1', 'wl', 'One-1'),
        ('ContributionTable', None, '12', 'wl', 'Dutch'),
        ('FormTable', None, '2', 'wl', 'eiland'),
        ('FormTable', None, '1', 'gn', 'a lot, '),
        ('FormTable', 'with_language', '2', 'wl', 'Dutch'),
        ('LanguageTable', None, 'Juang_SM', 'sd', lambda s: 'glottolog.org' not in s),
        ('LanguageTable', 'with_glottolog_link', 'Juang_SM', 'sd', 'glottolog.org'),
        ('ParameterTable', None, 'C', 'sd', 'Codes'),
        ('Source', None, 'Peterson2017', 'sd', 'Peterson, John. 2017.'),
        ('Source', None, '__all__', 'sd', '- '),
        ('Source', 'with_link', 'Peterson2017', 'sd', '[DOI:'),
    ]
)
def test_templates(Wordlist, Generic, StructureDataset, comp, query, oid, ds, expected):
    if ds == 'wl':
        dataset = Wordlist
    elif ds == 'gn':
        dataset = Generic
    else:
        dataset = StructureDataset
    res = render(
        '[]({}{}#cldf:{})'.format(comp, '?{}'.format(query) if query else '', oid), dataset)
    if callable(expected):
        assert expected(res)
    else:
        print(res)
        assert expected in res



import jinja2

import cldfviz

__all__ = ['render_jinja_template', 'TEMPLATE_DIR']

TEMPLATE_DIR = cldfviz.PKG_DIR / 'templates'


def render_jinja_template(path, **vars) -> str:
    loader = jinja2.FileSystemLoader(searchpath=[str(path.parent)])
    env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
    return env.get_template(path.name).render(**vars)



import argparse

from cldfviz import cli_util


def test_tree(tmp_path):
    nwk = tmp_path / 'test.nwk'
    nwk.write_text('((a,b),c)d;')
    parser = argparse.ArgumentParser()
    cli_util.add_tree(parser)
    args = parser.parse_args(['--tree', str(nwk)])
    res = cli_util.get_tree(args)
    assert res[0].name == 'd'



"""
Build an HTML file to display examples in a CLDF dataset.
"""
from pycldf.cli_util import add_dataset, get_dataset
from pycldf.sources import Sources
from clldutils.misc import nfilter

from cldfviz.cli_util import (
    add_open, write_output, add_jinja_template, add_language_filter, get_filtered_languages,
)
from cldfviz.media import get_objects_and_media, get_media_url
from cldfviz.template import render_jinja_template, TEMPLATE_DIR


def register(parser):
    add_dataset(parser)
    add_language_filter(parser)
    mod = __name__.split('.')[-1]
    add_jinja_template(parser, TEMPLATE_DIR / mod / '{}.html'.format(mod))
    add_open(parser)


def run(args):
    ds = get_dataset(args)
    valid_langs = get_filtered_languages(args, ds)

    examples = []
    for example, media in get_objects_and_media(
            ds, 'ExampleTable', 'exampleReference',
            filter=lambda e: (valid_langs is None) or e.cldf.languageReference in valid_langs):
        examples.append((
            example,
            nfilter(get_media_url(f) for f in media if f.mimetype.type == 'audio')))

    res = render_jinja_template(args.template, ds=ds, examples=examples, split_ref=Sources.parse)
    write_output(args, res)
